## file: Rcodes-chap06-statmodel.r
## usage: source('Rcodes-chap06-statmodel.r',echo=TRUE,max.deparse.length=Inf)
## required file: optdigits.tra
msg <- "------------------------ Type  <Return> to continue ------------------------: "
require(doParallel)                   # foreach(並列計算)
cl <- makeCluster(detectCores())      # クラスタの作成
registerDoParallel(cl)


## 光速度観測データ(morley)のプロット
data(morley)               # データ読み込み
dat <- morley[,3]+299000   # データを補正
hist(dat,breaks=seq(299600, 300100, by=20), col=5, ylim=c(0,17))   # プロット


readline(msg)
## 一様分布U[0,1]のパラメータ推定：最尤推定
x <- runif(10)   ## データ数 10
max(x)           ## 最尤推定
x <- runif(100)  ## データ数 100
max(x)           ## 最尤推定


readline(msg)
## 一様分布U[0,1]のパラメータ推定：不偏推定
n <- 10                 # データ数
x <- runif(n)           # データ生成
((n+1)/n) * max(x)      # 不偏推定
## 100セットのデータを生成
X <- matrix(runif(100*n),100,n)
mean(apply(X,1,max))           # 最尤推定の平均値: applyで行列Xの各行にmaxを適用
mean(((n+1)/n)*apply(X,1,max)) # 不偏推定の平均値


readline(msg)
## 最尤推定：表現型から遺伝子型の比率を推定
## 負の対数尤度を計算する関数
nlikelihood <- function(theta,n){
  a <- theta[1]
  b <- theta[2]
  o <- 1-a-b
  # 確率値の対数をそれぞれ計算
  logA  <- log(a^2+2*a*o)
  logB  <- log(b^2+2*b*o)
  logAB <- log(2*a*b)
  logO  <- log(o^2)
  # 負の対数尤度
  -sum(n*c(logA, logB, logAB, logO))
}
## 最尤推定量(mle)を求める関数
mle <- function(n){
  # 最適化の計算
  op <- optim(c(1/3,1/3),nlikelihood,n=n) 
  # 解を出力
  list(a=op$par[1],b=op$par[2],o=1-op$par[1]-op$par[2])
}
## source('bloodtype.r')   # プログラム読み込み


readline(msg)
## 表現型から遺伝子型の比率を推定： 例．A:40人，B:30人，AB:10人，O: 20人
mle(c(40,30,10,20))  


readline(msg)
## EMアルゴリズムで最尤推定：混合ベルヌーイ分布
## x:(データ数n, 次元d) 型行列. K:コンポーネント数. 
EM_mixBernoulli <- function(x,K=5,maxitr=1000,tol=1e-5){
  d <- ncol(x); n <- nrow(x)         # 次元dとデータ数n
  eps <- .Machine$double.eps         
  ## コンポーネント初期設定
  mu <- mean(x)                      
  p <- matrix(rbeta(K*d,shape1=mu,shape2=(1-mu)),K)
  q <- rep(1/K,K)                    # 混合確率の初期値
  ul <- Inf
  for(itr in 1:maxitr){              # EMアルゴリズム
    ## 多次元ベルヌーイ分布の確率を計算
    mp <- exp(log(p)%*%t(x)+log(1-p)%*%t(1-x))*q
    ## γ, q, p 更新．pmin, pmax で発散を防ぐ．
    gmm <- pmin(pmax(t(t(mp)/colSums(mp)),eps),1-eps)
    q <- pmin(pmax(rowSums(gmm)/n, eps),1-eps)
    p <- pmin(pmax((gmm%*%x)/(n*q),eps),1-eps)
    ## 負の対数尤度の上界
    uln <- -sum(gmm*((log(p)%*%t(x)+log(1-p)%*%t(1-x)+log(q))-log(gmm)))
    if(abs(ul-uln)<tol){           # 停止条件
      break
    }
    ul <- uln
  }
  BIC <- ul+0.5*(d*K+(K-1))*log(n) # BIC
  list(p=p,q=q,gamma=gmm,BIC=BIC)
}
## source('statmodel-em.r')


readline(msg)
## 例：optdigits.tra に適用
a <- read.table('optdigits.tra',sep=',')  # データ読込み
x <- as.matrix(a[,1:64]>8)      ## 16段階調の8以下を0とし，9以上を1に変換
y <- as.factor(a[,65])          ## 各画像データのラベル．混合モデルの推定には使わない
est <- EM_mixBernoulli(x,K=10)  ## 混合ベルヌーイ分布でデータの分布を推定:コンポーネント数は10
ec <- apply(est$gamma,2,which.max)  ## 各データがどのコンポーネントに属すかを計算 
table(y[which(ec==1)])   ## 第1コンポーネントに属すデータのラベルを表示
table(y[which(ec==2)])   ## 第2コンポーネントに属すデータのラベルを表示 
est$BIC  # BICの値


readline(msg)
## コンポーネント数 K(5-50) に対する BIC
Klist <- c(5:20,seq(22,50,by=2))
system.time(                              # 計算時間を測定
BIC_list <- foreach(K=Klist,.combine=c)%dopar%{ # 並列計算
  est <- EM_mixBernoulli(x,K=K,maxitr=1000,tol=1e-5)
  est$BIC
})
plot(Klist,BIC_list)          # プロット
stopCluster(cl)         # 並列計算を終了 
