## file: Rcodes-chap08-regression.r
## usage: source('Rcodes-chap08-regression.r',echo=TRUE,max.deparse.length=Inf)
msg <- "------------------------ Type  <Return> to continue ------------------------: "
require(car)       # UNデータ
require(MASS)      # rlm
require(glmnet)    # glmnet
require(kernlab)   # kernelMatrix
require(CVST)      # constructKRRLearner


## 最小二乗法
data(UN)       # データ読み込み
## データのプロット 
plot(UN[,2],UN[,1],log='xy',xlab=colnames(UN)[2],ylab=colnames(UN)[1])
logUN <- log(na.omit(UN))                   # データを対数で変換
res <- lm(infant.mortality~gdp,data=logUN)  # 最小二乗法
res                                         
y <- predict(res)                           # 予測値
lines(exp(cbind(logUN$gdp,y)),col=2,lwd=2)  # 回帰関数をプロット


readline(msg)
## 外れ値を含むデータに最小二乗法を適用
data(Davis)                        # データ読み込み
a <- lm(height~weight,data=Davis)  # 最小二乗法
a
readline(msg)
lm(height~weight,data=Davis[-12,]) # 最小二乗法(12番データを除く)

readline(msg)
## ロバスト回帰
b <- rlm(height~weight,data=Davis)   # rlmによるロバスト回帰
b
## データのプロット  
plot(Davis$weight,Davis$height,xlab='weight',ylab='height')
tx <- data.frame(weight=seq(40,160,l=100))       # 予測点
lines(tx$weight,predict(a,tx),col=2,lwd=3)       # 最小二乗法のプロット(実線)
lines(tx$weight,predict(b,tx),col=4,lwd=3,lty=2) # ロバスト推定のプロット(破線)


readline(msg)
## ロバスト回帰：rlm関数のオプションを psi.bisquare として推定   
rlm(height~weight,data=Davis,psi="psi.bisquare")


readline(msg)
## リッジ回帰
n <- 100                             # データ数
degree <- 8                          # 多項式モデルの次数
lambda_candidate <- 2^c(-12,-6,0,6)  # 正則化パラメータの候補
## データ生成
x <- runif(n,min=-2,max=2)
y <- sin(2*pi*x)/x + rnorm(n,sd=0.5)
## データ行列
Phi <- outer(x,0:degree,FUN="^")
IE  <- diag(ncol(Phi)); IE[1,1] <- 0
## パラメータ推定
par <- c()
for(lambda in lambda_candidate){
  par <- cbind(par, solve(t(Phi)%*%Phi+lambda*IE,t(Phi)%*%y))
}
## プロット
plot(x,y,lwd=2)                              # データ点プロット
x_test   <- seq(-2,2,l=100)                  # テスト点
Phi_test <- outer(x_test,0:degree,FUN="^")   # テスト点でのデータ行列
for (i in 1:ncol(par)){
  ypred <- c(Phi_test %*% par[,i])           # 予測値
  lines(x_test, ypred, lty=i, lwd=2, col=2)  # 予測値のプロット
}
legend("topright", paste("lambda=2^(",c(-12,-6,0,6),")",sep=""),lty=1:4,lwd=rep(2,4),col=rep(2,4))



readline(msg)
## リッジ回帰：glmnet 関数を使う
n <- 100           # データ数
degree <- 8        # 多項式モデルの次数
lambda <- 2^(-12)  # 正則化パラメータ
## データ生成
x <- runif(n,min=-2,max=2)
y <- sin(2*pi*x)/x + rnorm(n,sd=0.5)
## データ行列
Phi <- outer(x,1:degree,FUN="^")
## glmnetによるリッジ回帰
gl <- glmnet(Phi,y,alpha=0,family="gaussian",lambda=lambda/n)
gl


readline(msg)
## 推定した回帰関数をプロット
tx <- seq(-2,2,l=100)                 # テスト点
testPhi <- outer(tx,1:degree,FUN="^") # テスト点でのデータ行列
predy <- predict(gl,newx=testPhi)     # 予測値
## プロット
plot(x,y)
lines(tx,predy,lwd=2,col=2)


readline(msg)
## 10個の正則化パラメータに対して回帰関数を推定
gl <- glmnet(Phi[,-1],y,alpha=0,family="gaussian",nlambda=10)
gl$a0     ## それぞれの正則化パラメータで推定された定数項の値
gl$beta   ## それぞれの正則化パラメータで推定された係数パラメータ


readline(msg)
## 正則化パラメータを 10^(-6:1) としてリッジ回帰．結果をプロット
gl <- glmnet(Phi,y,alpha=0,family="gaussian",lambda=10^(-6:1))
tx <- seq(-2,2,l=100)                    # テスト点
testPhi <- outer(tx,1:degree,FUN="^")    # テスト点でのデータ行列
## 予測値：(テスト点の数, lambdaの候補数)の行列 
predy <- predict(gl,newx=testPhi)        
plot(x,y)                                # データ点のプロット
for(i in 1:ncol(predy)){
  lines(tx,predy[,i],col=2)              # 予測値のプロット
}


readline(msg)
## カーネル回帰分析
n <- 100                                 # データ数
## データ生成
x <- matrix(runif(n,min=-4,max=4))
y <- sin(pi*x)/x + rnorm(n,sd=0.5)
sigma  <- 3                              # カーネル幅パラメータの設定
K <- kernelMatrix(rbfdot(sigma=sigma),x) # ガウスカーネルのグラム行列
lambda <- 1                              # 正則化パラメータ設定
## カーネル回帰の極値条件に現れる行列とベクトル
extK <- rbind(cbind(rep(1,n), K+lambda*diag(n)), c(n, rowSums(K)))
exty <- c(y,sum(y))
## 推定パラメータを計算
par <- solve(extK, exty)
theta0 <- par[1]; alpha <- par[-1]
## プロット
predy <- theta0 + K %*% alpha
plot(x,y)
lines(cbind(x,predy)[order(x),],col=2,lwd=2)


readline(msg)
## CVSTパッケージを使ってカーネル回帰分析
dat <- constructData(x=x,y=c(y))          # データ生成
n <- getN(dat)                            # データ数の取得
## カーネルパラメータの推定 
par <- list(kernel="rbfdot",sigma=sigma,lambda=lambda/n)
KRR <- constructKRRLearner()              # オブジェクト生成
m <- KRR$learn(dat, par)                  # データから学習
predyCV <- KRR$predict(m, dat)            # 予測
plot(dat)                                 # プロット
lines(cbind(dat$x,predyCV)[order(dat$x),],col=2,lwd=2)
##
predy2  <- K %*% solve(K+diag(lambda,n),y)  # グラム行列から予測値を計算
mean(abs(predyCV-predy2))                   # 2種類の計算法の平均誤差

