## file: Rcodes-chap10-svm.r
## usage: source('Rcodes-chap10-svm.r',echo=TRUE,max.deparse.length=Inf)

msg <- "------------------------ Type  <Return> to continue ------------------------: "
require(kernlab)                  # ksvm など
require(mlbench)                  # mlbench.2dnormals を使う
require(doParallel)               # foreach(並列計算)
cl <- makeCluster(detectCores())  # クラスタの作成
registerDoParallel(cl)


## 判別境界のプロット
## sv: ksvm の出力オブジェクト
## dat=list(x,y): 多値トレーニングデータ
PlotDecisionBoundary <- function(sv,dat,len=100,main=NULL){
  x <- dat$x; cl <- as.numeric(dat$cl)
  r1 <- range(x[,1]); r2 <- range(x[,2])
  ## トレーニングデータのプロット  
  plot(data.frame(x),col=cl,pch=cl,xlim=r1,ylim=r2,lwd=2,cex=1.1,axes=TRUE,main=main)
  ## 格子点上でラベル予測
  x1 <- seq(r1[1],r1[2],l=len); x2 <- seq(r2[1],r2[2],l=len)
  X <- expand.grid(x1,x2)                                
  Y <- as.numeric(predict(sv,X)); Y <- matrix(Y,length(x1))  
  for(i in 1:max(Y)){  # 判別境界のプロット
    contour(x1,x2,Y==i,levels=0.5,lwd=2,lty=1,col=4,drawlabels=FALSE,add=TRUE)
  }
}


## SVM(線形カーネル)をmlbench.2dnormalsに適用
dat <- mlbench.2dnormals(200,cl=2,sd=1)                 # 200個の学習データを生成
## plot(dat, main="mlbench.2dnormalsで生成したデータ")    # データのプロット
x <- dat$x; y <- dat$classes             # x:入力の行列，y:ラベルのベクトル
## 線形判別器(vanilladot)で学習．結果を linsvm に代入
linsvm <- ksvm(x,y,type="C-svc",kernel="vanilladot")  
linsvm    # 結果を表示
## プロット
par(mfrow=c(1,1))
PlotDecisionBoundary(linsvm,dat,len=500,main="SVM(線形カーネル)")



readline(msg)
## テスト誤差を計算
tdat <- mlbench.2dnormals(1000,cl=2,sd=1)  ## テスト誤差の計算のため1000個のテストデータを生成
tx <- tdat$x; ty <- tdat$classes 
predy <- predict(linsvm,tx)                ## 学習した線形判別器linsvmでtxのラベルを予測
predy               # 予測ラベルの表示
mean(predy != ty)   # テスト誤差(近似値)



readline(msg)
## カーネル・サポートベクトルマシン
dat <- mlbench.spirals(300,cycles=1,sd=0.15)  # トレーニングデータ生成
testdat <- mlbench.spirals(1000,cycles=1,sd=0.15)  # テストデータ生成
readline(msg)
sv2 <- ksvm(dat$x,dat$c,kernel="polydot",kpar=list(degree=2))  # 2次多項式カーネルによる学習
sv2                                       # 学習結果の表示
mean(predict(sv2,testdat$x)!=testdat$c)   # テスト誤差の計算
readline(msg)
sv3 <- ksvm(dat$x,dat$c,kernel="polydot",kpar=list(degree=3))  # 3次多項式カーネルによる学習
sv3                                       # 学習結果の表示
mean(predict(sv3,testdat$x)!=testdat$c)   # テスト誤差の計算
## plot
par(mfrow=c(1,2))
PlotDecisionBoundary(sv2,dat,len=300,main="2次多項式カーネル")
PlotDecisionBoundary(sv3,dat,len=300,main="3次多項式カーネル")


readline(msg)
## ksvm の cross オプションで検証誤差を計算
## 例：x,y にデータを代入して以下を実行
x <- dat$x; y <- dat$c
sv <- ksvm(x,y,cross=10,kernel='rbfdot',kpar=list(sigma=1),C=1) # ガウスカーネルによる学習
cross(sv)      # 検証誤差の出力


readline(msg)
## 交差検証法の実装
## registerDoParallel(detectCores())   # コア数を取得
# モデルパラメータ mpar に対して交差検証法でテスト誤差を推定
# x,y: データ, K: K-重交差検証法
cvsvm <- function(x,y,K=5,kernel="rbfdot",mpar=list(kpar=list(sigma=1),C=1)){
  n <- nrow(x)  # データ数
  ## データをランダムにK分割
  idx <- sample(K,n,replace=TRUE) 
  ## 以下の %do% を %dopar% にすると並列実行
  cv_error <- foreach(l=1:K,.combine=c,.packages="kernlab")%dopar%{
    cvx <- x[idx!=l,]; cvy <- y[idx!=l]  # 交差検証法で学習に使うデータ
    sv <- ksvm(cvx,cvy,type="C-svc",kernel=kernel,kpar=mpar$kpar,C=mpar$C)
    pred <- predict(sv,x[idx==l,])       # 学習に使っていないデータのラベルを予測
    mean(pred != y[idx==l])              # 学習に使っていないデータに対する誤差
  }
  mean(cv_error)  # 出力値(検証誤差)
}


readline(msg)
## 交差検証法でモデルパラメータを選択
## source('cvsvm.r')
## データ生成
dat<-mlbench.spirals(200,cycles=1.2,sd=0.16)
x<-dat$x; y <- dat$classes  
cvK <- 5                                    # 5-重交差検証
sigma_candidate <- seq(0.01,20,length=30)   # sigmaの候補
pred_est <- foreach(sigma=sigma_candidate,.combine=c)%do%{     # 交差検証法で検証誤差を計算
  mpar <- list(kpar=list(sigma=sigma),C=1)
  cvsvm(x,y,K=cvK,kernel='rbfdot',mpar)
}
opt_sigma <- sigma_candidate[which.min(pred_est)]  # sigmaの選択
opt_sigma                                          # 最適なsigma
## sigmaに対する検証誤差のプロット
par(mfrow=c(1,1))
plot(sigma_candidate,pred_est,lwd=2, main="sigmaに対する検証誤差",xlab="sigma", ylab="検証誤差")      


readline(msg)
## 多値判別
G <- 8              # 8クラスの多値判別
train <- mlbench.2dnormals(500, cl=G,sd=0.8)    # トレーニングデータ
test  <- mlbench.2dnormals(1000,cl=G,sd=0.8)    # テストデータ
## ２値の組み合わせで予測 (多値判別に対するksvmのデフォルト)
## 線形カーネル
linsv <- ksvm(train$x,train$classes,kernel='vanilladot')
mean(predict(linsv,test$x)!=test$class)         # SVM(線形カーネル)のテスト誤差
## ガウスカーネル
rbfsv <- ksvm(train$x,train$classes,kernel='rbfdot')
mean(predict(rbfsv,test$x)!=test$class)         # SVM(ガウスカーネル)テスト誤差
## プロット
par(mfrow=c(1,2))
PlotDecisionBoundary(linsv,train,len=300,main="線形カーネル")
PlotDecisionBoundary(rbfsv,train,len=300,main="ガウスカーネル")



readline(msg)
## 手書き文字認識
## データ読込
a <- read.table('optdigits.tra',sep=',')
b <- read.table('optdigits.tes',sep=',')
x  <- as.matrix(a[,1:64]); y<-as.factor(a[,65])
tx <- as.matrix(b[,1:64]);ty<-as.factor(b[,65])
readline(msg)
## 線形カーネルで学習
linsv <- ksvm(x,y,scaled=F,kernel='vanilladot')    
predy <- predict(linsv,tx)                         # 予測ラベル
mean(predy!=ty)                                    # テスト誤差



readline(msg)
## ガウスカーネルによる学習
sigma <- sigest(x,scaled=F)[2]             # カーネル幅をsigestで設定 
sigma                                      # σの値
gsv <- ksvm(x,y,scaled=F,kernel='rbfdot',kpar=list(sigma=sigma))
predy <- predict(gsv,tx)           
mean(predy!=ty)                            # テスト誤差
## ガウスカーネルのカーネル幅を変更
sigma <- 0.01                      
gsv <- ksvm(x,y,scaled=F,kernel='rbfdot',kpar=list(sigma=sigma))
predy <- predict(gsv,tx)       
mean(predy!=ty)                            # テスト誤差

stopCluster(cl)

