## file: Rcodes-chap13-gaussprocess.r
## usage: source('Rcodes-chap13-gaussprocess.r',echo=TRUE,max.deparse.length=Inf)
msg <- "------------------------ Type  <Return> to continue ------------------------: "
require(kernlab)               # kernelMatrix, gausspr
require(mlbench)                # mlbench.spirals
require(rBayesianOptimization)  # BayesianOptimization


## ガウス過程モデルによる単回帰
n <- 30                  # データ数
theta <- 1               # θ=1
sd <- 0.5                # εの標準偏差
GPk <- function(x,z){    # k(x,z)を一般次元の入力に対して定義
  t(x)%*%z+sd^2*(all(x==z))
}
## データ生成
X   <- matrix(rnorm(n,sd=3))
Yob <- X%*%theta+matrix(rnorm(n,sd=sd))
newx <- matrix(seq(-3,3,length=100))        # 予測点
## カーネル関数の計算
Kob <- kernelMatrix(GPk,X)
kx  <- kernelMatrix(GPk,newx,X)
knewx <- kernelMatrix(GPk,newx,newx)
GPf   <- kx %*% solve(Kob,Yob)                      # ガウス過程による回帰関数の予測
GPv <- pmax(diag(knewx-kx %*% solve(Kob,t(kx))),0)  # 予測値の分散．
# プロット
plot(newx, GPf,type='l',lwd=2,main="回帰関数の推定と信頼区間")  # 推定した回帰関数
points(X,Yob)                                                           # データ点
# 信頼区間
lines(newx,GPf+sqrt(GPv),lty=2) 
lines(newx,GPf-sqrt(GPv),lty=2)


readline(msg)
## ガウス過程モデル(GP)による回帰分析
n <- 30                    # データ数
theta <- 1                 # θ=1
sd <- 0.5                  # εの標準偏差
## データ生成
X   <- matrix(rnorm(n,sd=3))
Yob <- X%*%theta+matrix(rnorm(n,sd=sd))
## ガウスカーネルによる予測
gp <- gausspr(X,Yob,kernel="rbfdot",variance.model=TRUE, var=sd^2)
newx <- matrix(seq(-3,3,length=100))           # 予測点
GPf  <- predict(gp,newx)                       # 回帰関数の推定
GPsd <- predict(gp,newx,type="sdeviation")     # 推定の標準偏差
## プロット
plot(newx, GPf, type='l',lwd=2, main="ガウス過程モデル(ガウスカーネル)で推定")
points(X,Yob)
lines(newx,GPf+GPsd,lty=2)
lines(newx,GPf-GPsd,lty=2)


readline(msg)
## ガウス過程モデル(GP)による判別分析
dat <- mlbench.spirals(200,cycles=1.2,sd=0.16)       # データ生成
gp <- gausspr(dat$x,dat$c,kernel="rbfdot")           # ガウス過程モデル(ガウスカーネル)
newx   <- matrix(seq(-1.2,1.2,length=100))           
newdat <- as.matrix(expand.grid(newx,newx))          # 2次元の予測点
GPf    <- predict(gp, newdat, type="prob")           # 確率の予測値
par(mfrow=c(1,2))                                    # プロット
plot(dat$x,col=dat$c,lwd=2,main="データ点")
image(matrix(GPf[,1],length(newx)),main="確率の予測値")


readline(msg)
## ベイズ最適化
dat <- mlbench.spirals(100,cycles=1.2,sd=0.16)    # データ生成
ksvm_CVaccuracy <- function(logC, logsig){        # 目的関数を定義
  sv <- ksvm(dat$x,dat$c,kernel ="rbfdot",
             kpar=list(sigma=10^logsig), C=10^logC, cross=5)
  list(Score=1-cross(sv),Pred=1-cross(sv))
}
sig <- sigest(dat$x)                              # カーネル幅の範囲をデータから決める
## logC と logsig の範囲を決め，リスト形式の変数 bounds に格納
bounds <- list(logC=c(-5,10),logsig=c(log(sig[1]/10,10),log(sig[3]*10,10)))


readline(msg)
## 獲得関数をUCBにしてベイズ最適化を実行
BOres <- BayesianOptimization(ksvm_CVaccuracy,bounds=bounds,init_points=10,n_iter=1,acq="ucb",kappa=1)
BOres$Best_Par   # 最適なパラメータ


readline(msg)
## 獲得関数を期待改善量にしてベイズ最適化を実行
BOres <- BayesianOptimization(ksvm_CVaccuracy,bounds=bounds,init_points=10, n_iter=1,acq="ei",eps=0.0)
BOres$Best_Par   # 最適なパラメータ
