## file: Rcodes-chap14-densityratio.r
## usage: source('Rcodes-chap14-densityratio.r',echo=TRUE,max.deparse.length=Inf)
msg <- "------------------------ Type  <Return> to continue ------------------------: "
require(kernlab)          # kernelMatrix
require(mlbench)          # LetterRecognition
require(Hotelling)        # hotelling.test
require(doParallel)       # foreach(並列計算)
cl <- makeCluster(detectCores())   # クラスタの作成
registerDoParallel(cl)


## 密度比推定の関数を定義
## ガウスカーネルによる密度比推定(係数αを計算): kernlab パッケージが必要
## de, nu: pde と pnu から得られたデータ行列
## sigma: カーネル幅パラメータ
## lambda: 正則化パラメータ
kernDR <- function(de,nu,
             sigma=sigest(rbind(de,nu))[2],
             lambda=min(nrow(de),nrow(nu))^(-0.9)){
  n <- nrow(de); m <- nrow(nu)        # データ数
  ## グラム行列の計算 
  Kdd <- kernelMatrix(rbfdot(sigma=sigma), de)
  Kdn <- kernelMatrix(rbfdot(sigma=sigma), de, nu)
  ## 係数パラメータαを求める
  alpha <- as.vector(solve(Kdd+n*lambda*diag(n),-rowMeans(Kdn)/lambda))
  ## 出力
  list(alpha=alpha,sigma=sigma,lambda=lambda,dat=list(de=de,nu=nu))
}
## 密度比の値を推定: kernlab パッケージが必要
## res: kernDRの出力, newdat: 予測点
predict.DR <- function(res,newdat){
  # カーネル密度比推定のそれぞれの項を計算  
  Wde <- kernelMult(rbfdot(sigma=res$sigma), newdat,res$dat$de,res$alpha)
  Wnu <- rowMeans(kernelMatrix(rbfdot(sigma=res$sigma), newdat,res$dat$nu))/res$lambda
  # 出力：予測点上での密度比の値
  c(pmax(Wde + Wnu,0))
}


readline(msg)
## 密度比を定義
## source('funcDR.r')                    # kernDR, predict.DR 
## データ設定 
n <- 100; m <- 200
de_mean <- 1   ; de_sd <- 0.8
nu_mean <- -0.5; nu_sd <- 1
newdat <- matrix(seq(-4,4,l=500))        # 予測点
## 真の密度比を計算
tnu <-  dnorm(newdat, mean=nu_mean, sd=nu_sd)
tde <- (dnorm(newdat, mean=de_mean, sd=de_sd)+tnu)/2
tw <- tnu/tde
par(mfrow=c(1,1))
plot(newdat,tw,lwd=2,col=1,type='l',ylim=c(0,2.5),main="密度比")     # 真の密度比をプロット 


readline(msg)
## 密度比をデータから推定
## データ生成
nu <- matrix(  rnorm(n, mean=nu_mean, sd=nu_sd))
de <- matrix(c(rnorm(n, mean=de_mean, sd=de_sd), rnorm(n, mean=nu_mean, sd=nu_sd)))
sigmas <- c(0.01, 0.1, 1)                                    # カーネル幅の候補
for(sigma in sigmas){
  res  <- kernDR(de,nu,sigma=sigma) #推定
  pred <- predict.DR(res,newdat)    #予測
  lines(newdat,pred,col=2,lwd=2,lty=2) #プロット
}
## データ点をプロット
points(cbind(nu,0.2),col=3)
points(cbind(de,0.1),col=4)


readline(msg)
## 密度比推定のための交差検証法
## データ nu, de, newdat は生成済み
K <- 5                                # K重交差検証法の K を設定
m <- nrow(nu);  n <- nrow(de);        # データ数
## カーネル幅パラメータの候補を生成
s <- sigest(rbind(de,nu))[1]
sigmas  <- exp(seq(log(s/100),log(s*10),l=20))   # カーネル幅パラメータの候補
# λ は１つ固定 (いくつかの候補を並べたベクトルでも可)
lambdas <- min(m,n)^(-0.9)
# モデルパラメータの候補
modelpars <- expand.grid(sigmas,lambdas)


readline(msg)
## モデルパラメータに対する検証誤差の計算
## nu, de データをそれぞれ Kグループに分ける
idx_nu <- rep(1:K,ceiling(m/K))[1:m] 
idx_de <- rep(1:K,ceiling(n/K))[1:n]
cvloss <- foreach(i=1:nrow(modelpars),.combine=c,.packages="kernlab")%dopar%{  # %dopar%とすると並列計算
  ## モデルパラメータ設定
  sigma <- modelpars[i,1]; lambda <- modelpars[i,2]
  tmpcvloss <- c()
  for(cvk in 1:K){ # 交差検証法の計算
    ## トレーニングデータ
    trnu <- nu[idx_nu!=cvk,,drop=FALSE]
    trde <- de[idx_de!=cvk,,drop=FALSE]
    ## テストデータ
    tenu <- nu[idx_nu==cvk,,drop=FALSE]
    tede <- de[idx_de==cvk,,drop=FALSE]
    ## 指定されたモデルパラメータで密度比を推定
    res <- kernDR(trde,trnu,sigma=sigma,lambda=lambda)
    ## 検証誤差の計算
    wnu <- predict.DR(res,tenu)
    wde <- predict.DR(res,tede)
    tmpcvloss <- c(tmpcvloss, mean(wde^2)/2-mean(wnu)) 
  }
  mean(tmpcvloss)
}
## 最適なモデルパラメータ
opt_sigma  <- modelpars[which.min(cvloss),1]
opt_lambda <- modelpars[which.min(cvloss),2]
print(c(opt_sigma, opt_lambda))    ## 最適なモデルパラメータ


readline(msg)
## プロット
par(mfrow=c(1,2))
plot(sigmas,cvloss,log='x')        ## sigma に対する検証誤差をプロット
## 最適なモデルパラメータで密度比を推定・予測
res <- kernDR(de,nu,sigma=opt_sigma,lambda=opt_lambda)
pred <- predict.DR(res,newdat)
## プロット
tnu <-  dnorm(newdat, mean=nu_mean, sd=nu_sd)
tde <- (dnorm(newdat, mean=de_mean, sd=de_sd)+tnu)/2
tw <- tnu/tde
plot(newdat,tw,lwd=2,col=1,type='l',ylim=c(0,2.4), xlab="x", ylab="密度比")    # 真の密度比
lines(newdat,pred,col=2,lwd=4,lty=2)                                          # 最適なモデルパラメータの結果
## その他のモデルパラメータで推定した結果をプロット
res <- kernDR(de,nu,sigma=sigmas[1],lambda=opt_lambda)
pred <- predict.DR(res,newdat)
lines(newdat,pred,col=5,lwd=2,lty=3)
## 
res <- kernDR(de,nu,sigma=sigmas[length(sigmas)],lambda=opt_lambda)
pred <- predict.DR(res,newdat)
lines(newdat,pred,col=5,lwd=2,lty=3)
## データ点のプロット
points(cbind(nu,0.2),col=3)
points(cbind(de,0.1),col=4)


readline(msg)
## 共変量シフトの下での回帰分析
f <- function(x){(x+2)*x*(x-3)}     # 真の関数
ntr <- 100; nte <- 100              # データ数
## 分布のパラメータ設定 
mtr <- -1.4; sdtr <- 0.7
mte <-  0.8; sdte <- 0.8
## トレーニングデータ生成
xtr <- matrix(rnorm(ntr,mean=mtr,sd=sdtr))
ytr <- f(xtr) + rnorm(ntr,sd=2)
## テストデータ生成
xte <- matrix(rnorm(ntr,mean=mte,sd=sdte))
yte <- f(xte) + rnorm(ntr,sd=2)
predw <- predict.DR(kernDR(xtr,xte),xtr)       # トレーニングデータ点上での密度比を推定
## 重み付き最小二乗法で回帰パラメータを推定
W <- sqrt(diag(predw))  
X <- cbind(1,xtr)
WX <- W %*%X; WY <- W %*% ytr;
estTheta <- c(solve(t(WX)%*%WX, t(WX)%*%WY)) # 重み付き最小二乗法の推定パラメータ


readline(msg)
## 共変量シフトの下での回帰分析：プロット
par(mfrow=c(1,1))
plot(xtr,ytr,col=2,xlim=c(-3,3),ylim=c(-18,17),lwd=1.5,xlab='x',ylab='y',main="共変量シフトの下での回帰分析")
points(xte,yte,col=4,lwd=1.5)
x <- seq(min(c(xtr,xte)),max(c(xtr,xte)),l=100); pX <- cbind(1,x)
lmest <- lm.fit(cbind(1,xtr),ytr)$coef          
lines(x, pX %*% lmest,  lwd=4, col=3, lty=2)   # 最小二乗法の結果をプロット
lmopt <- lm.fit(cbind(1,xte),yte)$coef          
lines(x, pX %*% lmopt, lwd=4, col=1, lty=1)    # 最適な１次式をプロット
lines(x, f(x),col=1,lwd=3,lty=2)               # 真の回帰関数
lines(x, pX %*% estTheta,lwd=7,col=6,lty=3)    # 重み付き最小二乗法
text(2, 13,expression(paste("最小二乗法")),col=3)
text(2,-13,expression(paste("推定目標")),col=1)
text(1.2,1.7,expression(paste("重み付き最小二乗法")),col=6)


readline(msg)
## 二標本検定
n <- 300; m <- 300                              # データ数
data(LetterRecognition); d <- LetterRecognition # データ読み込み
de <- as.matrix(d[d[,1]=='C',-1])
de <- de[sample(nrow(de),n),]                   # データ行列 de 生成
nu <- as.matrix(d[d[,1]=='G',-1])
nu <- nu[sample(nrow(nu),m),]                   # データ行列 nu 生成


readline(msg)
hotelling.test(de,nu)$pval     # ホテリングT^2検定のp値


readline(msg)
## 密度比推定を使った並べ替え検定
res  <- kernDR(de,nu)                          # 密度比の推定
L1distEst <- mean(abs(1-predict.DR(res,de)))   # L1距離の推定
## 並べ替え検定の繰り返し数を設定．時間が掛かるなら小さな値に設定
nperm <- 10000
system.time(                                # 計算時間を計測
permL1dist <- foreach(i=1:nperm,.combine=c,.packages="kernlab")%dopar%{# 並列計算
  ## データを合併して並べ替え
  idx <- which(sample(c(rep(0,nrow(nu)),rep(1,nrow(de))))==1)
  perm_de <- rbind(de,nu)[ idx,,drop=FALSE]
  perm_nu <- rbind(de,nu)[-idx,,drop=FALSE]
  ## L1距離を推定
  res <- kernDR(perm_de,perm_nu)            # 密度比の推定
  dr  <- predict.DR(res,perm_de)            # de上での密度比の値
  mean(abs(1-dr))                           # L1距離を推定
})
mean(L1distEst < permL1dist)                # 並べ替え検定による p値


readline(msg)
## スケーリングしたデータを検定
de <- scale(de)
nu <- scale(nu)

readline(msg)
## スケーリングしたデータに対するホテリングT^2検定
hotelling.test(de,nu)$pval     # ホテリングT^2検定のp値


readline(msg)
## スケーリングしたデータに対する並べ替え検定
res  <- kernDR(de,nu)                       # 密度比の推定
L1distEst <- mean(abs(1-predict.DR(res,de)))   # L1距離の推定
## 並べ替え検定の繰り返し数を設定．時間が掛かるなら小さな値に設定
nperm <- 10000
system.time(                                # 計算時間を計測
permL1dist <- foreach(i=1:nperm,.combine=c,.packages="kernlab")%dopar%{# 並列計算
  ## データを合併して並べ替え
  idx <- which(sample(c(rep(0,nrow(nu)),rep(1,nrow(de))))==1)
  perm_de <- rbind(de,nu)[ idx,,drop=FALSE]
  perm_nu <- rbind(de,nu)[-idx,,drop=FALSE]
  ## L1距離を推定
  res <- kernDR(perm_de,perm_nu)            # 密度比の推定
  dr  <- predict.DR(res,perm_de)               # de上での密度比の値
  mean(abs(1-dr))                           # L1距離を推定
})
mean(L1distEst < permL1dist)                # 並べ替え検定による p値
stopCluster(cl)                             # 並列計算を終了 

